@@ -356,6 +356,17 @@ def build_docker_image(be: BuildEnviron):  # pragma: no cover
         "--file", "release/docker/Dockerfile",
         "."
     ])
+    # smoke-test the newly built docker image
+    r = subprocess.run([
+        "docker",
+        "run",
+        "--rm",
+        be.docker_tag,
+        "mitmdump",
+        "--version",
+    ], check=True, capture_output=True)
+    print(r.stdout.decode())
+    assert "Mitmproxy: " in r.stdout.decode()
 
 
 def build_pyinstaller(be: BuildEnviron):  # pragma: no cover
